<?php

/*
 * Copyright (C) 2017 Champs-Libres <info@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\ActivityBundle\Export\Export;

use Chill\MainBundle\Export\ListInterface;
use Chill\ActivityBundle\Entity\ActivityReason;
use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Entity\Scope;
use Chill\ActivityBundle\Entity\ActivityType;
use Doctrine\ORM\Query\Expr;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Symfony\Component\Security\Core\Role\Role;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;
use Symfony\Component\Form\FormBuilderInterface;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Translation\TranslatorInterface;
use Symfony\Component\Validator\Constraints\Callback;
use Doctrine\ORM\Query;
use Chill\MainBundle\Export\FormatterInterface;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Validator\Context\ExecutionContextInterface;

/**
 * Create a list for all activities
 *
 * @author Julien Fastré <julien.fastre@champs-libres.coop>
 */
class ListActivity implements ListInterface
{

    /**
     *
     * @var EntityManagerInterface
     */
    protected $entityManager;

    /**
     *
     * @var TranslatorInterface
     */
    protected $translator;

    /**
     *
     * @var TranslatableStringHelper
     */
    protected $translatableStringHelper;

    protected $fields = array(
        'id',
        'date',
        'durationTime',
        'attendee',
        'list_reasons',
        'user_username',
        'circle_name',
        'type_name',
        'person_firstname',
        'person_lastname',
        'person_id'
    );

    public function __construct(
            EntityManagerInterface $em,
            TranslatorInterface $translator,
            TranslatableStringHelper $translatableStringHelper
            )
    {
        $this->entityManager = $em;
        $this->translator = $translator;
        $this->translatableStringHelper = $translatableStringHelper;
    }

    /**
     * {@inheritDoc}
     *
     * @param FormBuilderInterface $builder
     */
    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add('fields', ChoiceType::class, array(
            'multiple' => true,
            'expanded' => true,
            'choices' => array_combine($this->fields, $this->fields),
            'label'    => 'Fields to include in export',
            'constraints' => [new Callback(array(
                'callback' => function($selected, ExecutionContextInterface $context) {
                    if (count($selected) === 0) {
                        $context->buildViolation('You must select at least one element')
                            ->atPath('fields')
                            ->addViolation();
                    }
                }
            ))]
        ));

    }

    /**
     * {@inheritDoc}
     *
     * @return type
     */
    public function getAllowedFormattersTypes()
    {
        return array(FormatterInterface::TYPE_LIST);
    }

    public function getDescription()
    {
        return "List activities";
    }

    public function getLabels($key, array $values, $data)
    {
        switch ($key)
        {
            case 'date' :
                return function($value) {
                    if ($value === '_header') return 'date';

                    $date = \DateTime::createFromFormat('Y-m-d H:i:s', $value);

                    return $date->format('d-m-Y');
                };
            case 'attendee':
                return function($value) {
                    if ($value === '_header') return 'attendee';

                    return $value ? 1 : 0;
                };
            case 'list_reasons' :
                /* @var $activityReasonsRepository EntityRepository */
                $activityRepository = $this->entityManager
                    ->getRepository('ChillActivityBundle:Activity');

                return function($value) use ($activityRepository) {
                    if ($value === '_header') return 'activity reasons';

                    $activity = $activityRepository
                            ->find($value);

                    return implode(", ", array_map(function(ActivityReason $r) {

                        return '"'.
                                $this->translatableStringHelper->localize($r->getCategory()->getName())
                                .' > '.
                                $this->translatableStringHelper->localize($r->getName())
                                .'"';
                    }, $activity->getReasons()->toArray()));
                };
            case 'circle_name' :
                return function($value) {
                    if ($value === '_header') return 'circle';

                    return $this->translatableStringHelper
                            ->localize(json_decode($value, true));
                };
            case 'type_name' :
                return function($value) {
                    if ($value === '_header') return 'activity type';

                    return $this->translatableStringHelper
                            ->localize(json_decode($value, true));
                };
            default:
                return function($value) use ($key) {
                    if ($value === '_header') return $key;

                    return $value;
                };
        }
    }

    public function getQueryKeys($data)
    {
        return $data['fields'];
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(Query::HYDRATE_SCALAR);
    }

    public function getTitle()
    {
        return 'List activities';
    }

    public function getType()
    {
        return 'activity';
    }

    public function initiateQuery(array $requiredModifiers, array $acl, array $data = array())
    {
        $centers = array_map(function($el) { return $el['center']; }, $acl);

        // throw an error if any fields are present
        if (!\array_key_exists('fields', $data)) {
            throw new \Doctrine\DBAL\Exception\InvalidArgumentException("any fields "
                . "have been checked");
        }

        $qb = $this->entityManager->createQueryBuilder();

        $qb
                ->from('ChillActivityBundle:Activity', 'activity')
                ->join('activity.person', 'person')
                ->join('person.center', 'center')
                ->andWhere('center IN (:authorized_centers)')
                ->setParameter('authorized_centers', $centers);
                ;

        foreach ($this->fields as $f) {
            if (in_array($f, $data['fields'])) {
                switch ($f) {
                    case 'id':
                        $qb->addSelect('activity.id AS id');
                        break;
                    case 'person_firstname':
                        $qb->addSelect('person.firstName AS person_firstname');
                        break;
                    case 'person_lastname':
                        $qb->addSelect('person.lastName AS person_lastname');
                        break;
                    case 'person_id':
                        $qb->addSelect('person.id AS person_id');
                        break;
                    case 'user_username':
                        $qb->join('activity.user', 'user');
                        $qb->addSelect('user.username AS user_username');
                        break;
                    case 'circle_name':
                        $qb->join('activity.scope', 'circle');
                        $qb->addSelect('circle.name AS circle_name');
                        break;
                    case 'type_name':
                        $qb->join('activity.type', 'type');
                        $qb->addSelect('type.name AS type_name');
                        break;
                    case 'list_reasons':
                        // this is a trick... The reasons is filled with the
                        // activity id which will be used to load reasons
                        $qb->addSelect('activity.id AS list_reasons');
                        break;
                    default:
                        $qb->addSelect(sprintf('activity.%s as %s', $f, $f));
                        break;
                }

            }
        }



        return $qb;
    }

    public function requiredRole()
    {
        return new Role(ActivityStatsVoter::LISTS);
    }

    public function supportsModifiers()
    {
        return array('activity', 'person');
    }

}
