<?php

/*
 * Copyright (C) 2018 Champs-Libres <info@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\ActivityBundle\Export\Filter;

use Chill\MainBundle\Export\FilterInterface;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Doctrine\ORM\Query\Expr;
use Symfony\Component\Security\Core\Role\Role;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\Query\Expr\Join;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Chill\MainBundle\Export\ExportElementValidatedInterface;
use Chill\ActivityBundle\Entity\ActivityType;

/**
 * 
 *
 */
class ActivityTypeFilter implements FilterInterface,
   ExportElementValidatedInterface
{
    /**
     *
     * @var TranslatableStringHelper
     */
    protected $translatableStringHelper;
    
    /**
     * The repository for activity reasons
     *
     * @var EntityRepository
     */
    protected $typeRepository;
    
    public function __construct(
            TranslatableStringHelper $helper,
            EntityRepository $typeRepository
    ) {
        $this->translatableStringHelper = $helper;
        $this->typeRepository = $typeRepository;
    }
    
    
    public function alterQuery(QueryBuilder $qb, $data)
    {
        $where = $qb->getDQLPart('where');
        $clause = $qb->expr()->in('activity.type', ':selected_activity_types');

        if ($where instanceof Expr\Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }
        
        $qb->add('where', $where);
        $qb->setParameter('selected_activity_types', $data['types']);
    }
    
    /**
     * Check if a join between Activity and Reason is already defined
     * 
     * @param Join[] $joins
     * @return boolean
     */
    private function checkJoinAlreadyDefined(array $joins, $alias)
    {
        foreach ($joins as $join) {
            if ($join->getAlias() === $alias) {
                return true;
            }
        }
        
        return false;
    }

    public function applyOn()
    {
        return 'activity';
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        //create a local copy of translatableStringHelper
        $helper = $this->translatableStringHelper;
        
        $builder->add('types', EntityType::class, array(
            'class' => ActivityType::class,
            'choice_label' => function (ActivityType $type) use ($helper) {
                return $helper->localize($type->getName());
            },
            'multiple' => true,
            'expanded' => false
        ));
    }
    
    public function validateForm($data, ExecutionContextInterface $context)
    {
        if ($data['types'] === null || count($data['types']) === 0) {
            $context->buildViolation("At least one type must be choosen")
                ->addViolation();
        }
    }

        public function getTitle()
    {
        return 'Filter by activity type';
    }
    
    public function addRole()
    {
        return new Role(ActivityStatsVoter::STATS);
    }
    
    public function describeAction($data, $format = 'string')
    {
        // collect all the reasons'name used in this filter in one array
        $reasonsNames = array_map(
                function(ActivityType $t) {
                    return "\"".$this->translatableStringHelper->localize($t->getName())."\"";
                },
                $this->typeRepository->findBy(array('id' => $data['types']->toArray()))
                );
        
        return array("Filtered by activity type: only %list%", 
            ["%list%" => implode(", ", $reasonsNames)]);
    }
}
