<?php

/*
 * Copyright (C) 2017 Champs-Libres <info@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\ActivityBundle\Export\Filter;

use Chill\MainBundle\Export\FilterInterface;
use Symfony\Component\Form\FormEvent;
use Symfony\Component\Form\FormEvents;
use Symfony\Component\Form\Extension\Core\Type\DateType;
use Symfony\Component\Form\FormError;
use Chill\MainBundle\Form\Type\Export\FilterType;
use Doctrine\ORM\Query\Expr;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Chill\ActivityBundle\Entity\ActivityReason;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\EntityManager;
use Chill\PersonBundle\Export\Declarations;
use Symfony\Component\Translation\TranslatorInterface;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Chill\MainBundle\Export\ExportElementValidatedInterface;

/**
 * 
 *
 * @author Julien Fastré <julien.fastre@champs-libres.coop>
 */
class PersonHavingActivityBetweenDateFilter implements FilterInterface, 
    ExportElementValidatedInterface
{
    
    /**
     *
     * @var TranslatableStringHelper
     */
    protected $translatableStringHelper;
    
    /**
     *
     * @var EntityRepository
     */
    protected $activityReasonRepository;
    
    /**
     *
     * @var TranslatorInterface 
     */
    protected $translator;
    
    public function __construct(
            TranslatableStringHelper $translatableStringHelper,
            EntityRepository $activityReasonRepository,
            TranslatorInterface $translator
    ) {
        $this->translatableStringHelper = $translatableStringHelper;
        $this->activityReasonRepository = $activityReasonRepository;
        $this->translator               = $translator;
    }


    public function addRole()
    {
        return null;
    }

    public function alterQuery(\Doctrine\ORM\QueryBuilder $qb, $data)
    {
        // create a query for activity
        $sqb = $qb->getEntityManager()->createQueryBuilder();
        $sqb->select("person_person_having_activity.id")
                ->from("ChillActivityBundle:Activity", "activity_person_having_activity")
                ->join("activity_person_having_activity.person", "person_person_having_activity")
                ;
        // add clause between date
        $sqb->where("activity_person_having_activity.date BETWEEN "
                . ":person_having_activity_between_date_from"
                . " AND "
                . ":person_having_activity_between_date_to");
        // add clause activity reason
        $sqb->join('activity_person_having_activity.reasons', 
                'reasons_person_having_activity');
        $sqb->andWhere(
                $sqb->expr()->in(
                    'reasons_person_having_activity',
                    ":person_having_activity_reasons")
                );
        
        $where = $qb->getDQLPart('where');
        $clause = $qb->expr()->in('person.id', $sqb->getDQL());

        if ($where instanceof Expr\Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }
        
        $qb->add('where', $where);
        $qb->setParameter('person_having_activity_between_date_from', 
                $data['date_from']);
        $qb->setParameter('person_having_activity_between_date_to', 
                $data['date_to']);
        $qb->setParameter('person_having_activity_reasons', $data['reasons']);
    }

    public function applyOn()
    {
        return Declarations::PERSON_IMPLIED_IN;
    }

    public function buildForm(\Symfony\Component\Form\FormBuilderInterface $builder)
    {
        $builder->add('date_from', DateType::class, array(
            'label' => "Implied in an activity after this date",
            'data'  => new \DateTime(),
            'attr'  => array('class' => 'datepicker'),
            'widget'=> 'single_text',
            'format' => 'dd-MM-yyyy',
        ));
        
        $builder->add('date_to', DateType::class, array(
            'label' => "Implied in an activity before this date",
            'data'  => new \DateTime(),
            'attr'  => array('class' => 'datepicker'),
            'widget'=> 'single_text',
            'format' => 'dd-MM-yyyy',
        ));
        
        $builder->add('reasons', EntityType::class, array(
            'class' => 'ChillActivityBundle:ActivityReason',
            'choice_label' => function (ActivityReason $reason) {
                return $this->translatableStringHelper
                        ->localize($reason->getName());
            },
            'group_by' => function(ActivityReason $reason) {
                return $this->translatableStringHelper
                        ->localize($reason->getCategory()->getName());
            },
            'data' => $this->activityReasonRepository->findAll(),
            'multiple' => true,
            'expanded' => false,
            'label'    => "Activity reasons for those activities"
        ));
        
        $builder->addEventListener(FormEvents::POST_SUBMIT, function(FormEvent $event) {
            /* @var $filterForm \Symfony\Component\Form\FormInterface */
            $filterForm = $event->getForm()->getParent();
            $enabled = $filterForm->get(FilterType::ENABLED_FIELD)->getData();
            
            if ($enabled === true) {
                // if the filter is enabled, add some validation
                $form      = $event->getForm();
                $date_from = $form->get('date_from')->getData();
                $date_to   = $form->get('date_to')->getData();
                
                // check that fields are not empty
                if ($date_from === null) {
                    $form->get('date_from')->addError(new FormError(
                        $this->translator->trans('This field '
                        . 'should not be empty')));
                }
                if ($date_to   === null) {
                    $form->get('date_to')->addError(new FormError(
                        $this->translator->trans('This field '
                        . 'should not be empty')));
                } 
                
                // check that date_from is before date_to
                if (
                    ($date_from !== null && $date_to !== null)
                    &&
                    $date_from >= $date_to
                ) {
                    $form->get('date_to')->addError(new FormError(
                        $this->translator->trans('This date '
                        . 'should be after the date given in "Implied in an '
                        . 'activity after this date" field')));
                }
            }
        });
    }
    
    public function validateForm($data, ExecutionContextInterface $context)
    {
        if ($data['reasons'] === null || count($data['reasons']) === 0) {
            $context->buildViolation("At least one reason must be choosen")
                ->addViolation();
        }
    }

    public function describeAction($data, $format = 'string')
    {
        return array(
            "Filtered by person having an activity between %date_from% and "
            . "%date_to% with reasons %reasons_name%",
            array(
                "%date_from%" => $data['date_from']->format('d-m-Y'),
                '%date_to%'   => $data['date_to']->format('d-m-Y'),
                "%reasons_name%"  => implode(", ", array_map(
                        function (ActivityReason $r) {
                            return '"'.$this->translatableStringHelper->
                                    localize($r->getName()).'"';
                        },
                        $data['reasons']))
            ));
    }

    public function getTitle()
    {
        return "Filtered by person having an activity in a period";
    }

}
