<?php

/*
 * Chill is a software for social workers
 *
 * Copyright (C) 2015, Champs Libres Cooperative SCRLFS,
 * <http://www.champs-libres.coop>, <info@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\PersonBundle\Controller;

use Chill\PersonBundle\Privacy\PrivacyEvent;
use Psr\Log\LoggerInterface;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Form\PersonType;
use Chill\PersonBundle\Form\CreationPersonType;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Security\Core\Role\Role;
use Chill\PersonBundle\Security\Authorization\PersonVoter;
use Chill\PersonBundle\Search\SimilarPersonMatcher;
use Symfony\Component\Translation\TranslatorInterface;
use Chill\MainBundle\Search\SearchProvider;
use Chill\PersonBundle\Repository\PersonRepository;
use Chill\PersonBundle\Config\ConfigPersonAltNamesHelper;
use Chill\PersonBundle\Repository\PersonNotDuplicateRepository;
use Symfony\Component\Validator\Validator\ValidatorInterface;
use Doctrine\ORM\EntityManagerInterface;

final class PersonController extends AbstractController
{
    /**
     * @var SimilarPersonMatcher
     */
    protected $similarPersonMatcher;

    /**
     * @var TranslatorInterface
     */
    protected $translator;

    /**
     * @var EventDispatcherInterface
     */
    protected $eventDispatcher;

    /**
     * @var PersonRepository;
     */
    protected $personRepository;

    /**
     * @var ConfigPersonAltNamesHelper
     */
    protected $configPersonAltNameHelper;

    /**
     * @var EntityManagerInterface
     */
    private $em;

    /**
     * @var \Psr\Log\LoggerInterface
     */
    private $logger;

    /**
     * @var ValidatorInterface
     */
    private $validator;

    public function __construct(
        SimilarPersonMatcher $similarPersonMatcher,
        TranslatorInterface $translator,
        EventDispatcherInterface $eventDispatcher,
        PersonRepository $personRepository,
        ConfigPersonAltNamesHelper $configPersonAltNameHelper,
        LoggerInterface $logger,
        ValidatorInterface $validator,
        EntityManagerInterface $em
    ) {
        $this->similarPersonMatcher = $similarPersonMatcher;
        $this->translator = $translator;
        $this->eventDispatcher = $eventDispatcher;
        $this->configPersonAltNameHelper = $configPersonAltNameHelper;
        $this->personRepository = $personRepository;
        $this->logger = $logger;
        $this->validator = $validator;
        $this->em = $em;
  }

    public function getCFGroup()
    {
        $cFGroup = null;

        $cFDefaultGroup  = $this->em->getRepository("ChillCustomFieldsBundle:CustomFieldsDefaultGroup")
            ->findOneByEntity("Chill\PersonBundle\Entity\Person");

        if($cFDefaultGroup) {
            $cFGroup = $cFDefaultGroup->getCustomFieldsGroup();
        }

        return $cFGroup;
    }

    public function viewAction($person_id)
    {
        $person = $this->_getPerson($person_id);

        if ($person === null) {
            throw $this->createNotFoundException("Person with id $person_id not"
                    . " found on this server");
        }

        $this->denyAccessUnlessGranted('CHILL_PERSON_SEE', $person,
                "You are not allowed to see this person.");

        $event = new PrivacyEvent($person);
        $this->eventDispatcher->dispatch(PrivacyEvent::PERSON_PRIVACY_EVENT, $event);

        return $this->render('ChillPersonBundle:Person:view.html.twig',
            array(
                "person" => $person,
                "cFGroup" => $this->getCFGroup(),
                "alt_names" => $this->configPersonAltNameHelper->getChoices(),
            ));
    }

    public function editAction($person_id)
    {
        $person = $this->_getPerson($person_id);

        if ($person === null) {
            return $this->createNotFoundException();
        }

        $this->denyAccessUnlessGranted('CHILL_PERSON_UPDATE', $person,
                'You are not allowed to edit this person');

        $form = $this->createForm(PersonType::class, $person,
            array(
                "action" => $this->generateUrl('chill_person_general_update',
                    array("person_id" => $person_id)),
                "cFGroup" => $this->getCFGroup()
            )
        );

        return $this->render('ChillPersonBundle:Person:edit.html.twig',
            array('person' => $person, 'form' => $form->createView()));
    }

    public function updateAction($person_id, Request $request)
    {
        $person = $this->_getPerson($person_id);

        if ($person === null) {
            return $this->createNotFoundException();
        }

        $this->denyAccessUnlessGranted('CHILL_PERSON_UPDATE', $person,
                'You are not allowed to edit this person');

        $form = $this->createForm(PersonType::class, $person,
            array("cFGroup" => $this->getCFGroup()));

        if ($request->getMethod() === 'POST') {
            $form->handleRequest($request);

            if ( ! $form->isValid() ) {
                $this->get('session')
                        ->getFlashBag()->add('error', $this->translator
                            ->trans('This form contains errors'));

                return $this->render('ChillPersonBundle:Person:edit.html.twig',
                        array('person' => $person,
                            'form' => $form->createView()));
            }

            $this->get('session')->getFlashBag()
                    ->add('success',
                            $this->get('translator')
                            ->trans('The person data has been updated')
                            );

            $this->em->flush();

            $url = $this->generateUrl('chill_person_view', array(
                'person_id' => $person->getId()
            ));

            return $this->redirect($url);
        }
    }

    public function newAction()
    {
        // this is a dummy default center.
        $defaultCenter = $this->get('security.token_storage')
                        ->getToken()
                        ->getUser()
                        ->getGroupCenters()[0]
                        ->getCenter();

        $person = (new Person(new \DateTime('now')))
                ->setCenter($defaultCenter);

        $form = $this->createForm(
                CreationPersonType::class,
                $person,
                array(
                   'action' => $this->generateUrl('chill_person_review'),
                   'form_status' => CreationPersonType::FORM_NOT_REVIEWED
                      ));

        return $this->_renderNewForm($form);
    }

    private function _renderNewForm($form)
    {
        return $this->render('ChillPersonBundle:Person:create.html.twig',
            array(
                'form' => $form->createView()
            ));
    }

    /**
     *
     * @param type $form
     * @return \Chill\PersonBundle\Entity\Person
     */
    private function _bindCreationForm($form)
    {
        /**
         * @var Person
         */
        $person = $form->getData();

        $periods = $person->getAccompanyingPeriodsOrdered();
        $period = $periods[0];
        $period->setOpeningDate($form['creation_date']->getData());
//        $person = new Person($form['creation_date']->getData());
//
//        $person->setFirstName($form['firstName']->getData())
//                ->setLastName($form['lastName']->getData())
//                ->setGender($form['gender']->getData())
//                ->setBirthdate($form['birthdate']->getData())
//                ->setCenter($form['center']->getData())
//                ;

        return $person;
    }

    /**
     *
     * @param \Chill\PersonBundle\Entity\Person $person
     * @return \Symfony\Component\Validator\ConstraintViolationListInterface
     */
    private function _validatePersonAndAccompanyingPeriod(Person $person)
    {
        $errors = $this->validator
            ->validate($person, null, array('creation'));

        //validate accompanying periods
        $periods = $person->getAccompanyingPeriods();

        foreach ($periods as $period) {
            $period_errors = $this->validator
                ->validate($period);

            //group errors :
            foreach($period_errors as $error) {
                $errors->add($error);
            }
        }

        return $errors;
    }

    public function reviewAction(Request $request, PersonNotDuplicateRepository $personNotDuplicateRepository)
    {
        if ($request->getMethod() !== 'POST') {
            $r = new Response("You must send something to review the creation of a new Person");
            $r->setStatusCode(400);
            return $r;
        }

        $form = $this->createForm(
                CreationPersonType::class,
                new Person(),
                array(
                   'action' => $this->generateUrl('chill_person_create'),
                   'form_status' => CreationPersonType::FORM_BEING_REVIEWED
                      ));

        $form->handleRequest($request);

        $person = $this->_bindCreationForm($form);

        $errors = $this->_validatePersonAndAccompanyingPeriod($person);
        $this->logger->info(sprintf('Person created with %d errors ', count($errors)));

        if ($errors->count() > 0) {
            $this->logger->info('The created person has errors');
            $flashBag = $this->get('session')->getFlashBag();
            $translator = $this->get('translator');

            $flashBag->add('error', $translator->trans('The person data are not valid'));

            foreach($errors as $error) {
                $flashBag->add('info', $error->getMessage());
            }

            $form = $this->createForm(
                CreationPersonType::class,
                $person,
                array(
                   'action' => $this->generateUrl('chill_person_review'),
                   'form_status' => CreationPersonType::FORM_NOT_REVIEWED
                      ));

            $form->handleRequest($request);

            return $this->_renderNewForm($form);
        } else {
            $this->logger->info('Person created without errors');
        }

        $this->em->persist($person);

        $alternatePersons = $this->similarPersonMatcher->matchPerson($person, $personNotDuplicateRepository);

        if (count($alternatePersons) === 0) {
            return $this->forward('ChillPersonBundle:Person:create');
        }

        $this->get('session')->getFlashBag()->add('info',
                $this->get('translator')->trans(
                        '%nb% person with similar name. Please verify that this is a new person',
                        array('%nb%' => count($alternatePersons)))
                );

        return $this->render('ChillPersonBundle:Person:create_review.html.twig',
            array(
                'person' => $person,
                'alternatePersons' => $alternatePersons,
                'firstName' => $form['firstName']->getData(),
                'lastName' => $form['lastName']->getData(),
                'birthdate' => $form['birthdate']->getData(),
                'gender' => $form['gender']->getData(),
                'creation_date' => $form['creation_date']->getData(),
                'form' => $form->createView()));
    }

    public function createAction(Request $request)
    {

        if ($request->getMethod() !== 'POST') {
            $r = new Response('You must send something to create a person !');
            $r->setStatusCode(400);
            return $r;
        }

        $form = $this->createForm(CreationPersonType::class, null, array(
           'form_status' => CreationPersonType::FORM_REVIEWED
        ));

        $form->handleRequest($request);

        $person = $this->_bindCreationForm($form);

        $errors = $this->_validatePersonAndAccompanyingPeriod($person);

        $this->denyAccessUnlessGranted('CHILL_PERSON_CREATE', $person,
                'You are not allowed to create this person');

        if ($errors->count() ===  0) {
            $this->em->persist($person);

            $this->em->flush();

            return $this->redirect($this->generateUrl('chill_person_general_edit',
                    array('person_id' => $person->getId())));
        } else {
            $text = "this should not happen if you reviewed your submission\n";
            foreach ($errors as $error) {
                $text .= $error->getMessage()."\n";
            }
            $r = new Response($text);
            $r->setStatusCode(400);
            return $r;
        }
    }

    /**
     * easy getting a person by his id
     * @return \Chill\PersonBundle\Entity\Person
     */
    private function _getPerson($id)
    {
        $person = $this->personRepository->find($id);

        return $person;
    }
}
